from collections import namedtuple
from typing import Dict, List, Union

import ee
import ipywidgets as widgets
import pandas as pd
import plotly.graph_objects as go

from sankee import sampling, utils

SankeyParameters = namedtuple(
    "SankeyParameters",
    [
        "node_labels",
        "link_labels",
        "node_palette",
        "link_palette",
        "label",
        "source",
        "target",
        "value",
    ],
)


def sankify(
    image_list: List[ee.Image],
    band: str,
    labels: Dict[int, str],
    palette: Dict[int, str],
    region: Union[None, ee.Geometry] = None,
    label_list: Union[None, List[str]] = None,
    max_classes: Union[None, int] = None,
    n: int = 500,
    title: Union[None, str] = None,
    scale: Union[None, int] = None,
    seed: int = 0,
    label_type: str = "class",
) -> go.Figure:
    """
    Generate an interactive Sankey plot showing land cover change over time from a series of images.

    Parameters
    ----------
    image_list : List[ee.Image]
        An ordered list of images representing a time series of classified data. Each image will be
        sampled to generate the Sankey plot. Any length of list is allowed, but lists with more than
        3 or 4 images may produce unusable plots.
    band : str
        The name of the band in all images of image_list that contains classified data.
    labels : dict
        The labels associated with each value of all images in image_list. Any values not defined
        in the labels will be dropped from the sampled data.
    palette : dict
        The colors associated with each value of all images in image_list.
    region : ee.Geometry, default None
        A region to generate samples within. The region must overlap all images. If none is
        provided, the geometry of the first image will be used. For this to work, images must be
        bounded.
    label_list : List[str], default None
        An ordered list of labels corresponding to the images. The list must be the same length as
        image_list. If none is provided, sequential numeric labels will be automatically assigned
        starting at 0. Labels are displayed on-hover on the Sankey nodes.
    max_classes : int, default None
        If a value is provided, small classes will be removed until max_classes remain. Class size
        is calculated based on total times sampled in the time series.
    n : int, default 500
        The number of sample points to randomly generate for characterizing all images. More samples
        will provide more representative data but will take longer to process.
    title : str, default None
        An optional title that will be displayed above the Sankey plot.
    scale : int, default None
        The scale in image units to perform sampling at. If none is provided, GEE will attempt to
        use the image's nominal scale, which may cause errors depending on the image projection.
    seed : int, default 0
        The seed value used to generate repeatable results during random sampling.
    label_type : str, default "class"
        The type of label to display for each link, one of "class", "percent", or "count". Selecting
        "class" will use the class label, "percent" will use the proportion of sampled pixels in each
        class, and "count" will use the number of sampled pixels in each class.

    Returns
    -------
    plotly.graph_objs._figure.Figure
        An interactive Sankey plot.
    """
    if region is None:
        region = image_list[0].geometry()

    label_list = label_list if label_list is not None else list(range(len(image_list)))
    label_list = [str(label) for label in label_list]
    if len(label_list) != len(image_list):
        raise ValueError("The number of labels must match the number of images.")
    if len(set(label_list)) != len(label_list):
        raise ValueError("All labels in the `label_list` must be unique.")

    data, samples = sampling.generate_sample_data(
        image_list=image_list,
        image_labels=label_list,
        band=band,
        scale=scale,
        include=list(labels.keys()),
        max_classes=max_classes,
        region=region,
        n=n,
        seed=seed,
    )

    return SankeyPlot(
        data=data,
        labels=labels,
        palette=palette,
        title=title,
        samples=samples,
        label_type=label_type,
    )


class SankeyPlot(widgets.DOMWidget):
    def __init__(
        self,
        data: pd.DataFrame,
        labels: Dict[int, str],
        palette: Dict[int, str],
        title: str,
        samples: ee.FeatureCollection,
        label_type: str,
    ):
        self.data = data
        self.labels = labels
        self.palette = palette
        self.title = title
        self.samples = samples
        self.label_type = label_type

        self.hide = []
        # Initialized by `self.generate_plot`
        self.df = None
        self.plot = self.generate_plot()
        self.gui = self.generate_gui()

    def get_sorted_classes(self) -> pd.Series:
        """Return all unique class values, sorted by the total number of observations."""
        start_count = (
            self.df.groupby("source")
            .mean()
            .reset_index()[["source", "total"]]
            .rename(columns={"source": "class", "total": "count"})
        )
        end_count = (
            self.df.groupby("target")
            .sum()
            .reset_index()[["target", "changed"]]
            .rename(columns={"target": "class", "changed": "count"})
        )
        total_count = pd.concat([start_count, end_count]).groupby("class").sum().reset_index()

        return total_count.sort_values(by="count", ascending=False)["class"].reset_index(drop=True)

    def get_active_classes(self) -> pd.Series:
        """Return all unique active, visibile class values after filtering."""
        return self.df[["source", "target"]].melt().value.unique()

    def generate_plot_parameters(self) -> SankeyParameters:
        """Generate Sankey plot parameters from a formatted, cleaned dataframe"""
        df = self.df.copy()

        source_df = df[["source", "source_year"]].rename(
            columns={"source": "class", "source_year": "year"}
        )
        target_df = df[["target", "target_year"]].rename(
            columns={"target": "class", "target_year": "year"}
        )
        all_classes = pd.concat([source_df, target_df])

        all_classes = all_classes.drop_duplicates().reset_index(drop=True)
        all_classes["color"] = all_classes["class"].apply(lambda k: self.palette[k]).tolist()
        all_classes["id"] = all_classes.groupby(["year", "class"], sort=False).ngroup()

        # Join the sequential class-year IDs to the dataframe
        df["source_id"] = pd.merge(
            left=df,
            right=all_classes,
            how="left",
            left_on=["source_year", "source"],
            right_on=["year", "class"],
        )["id"]
        df["target_id"] = pd.merge(
            left=df,
            right=all_classes,
            how="left",
            left_on=["target_year", "target"],
            right_on=["year", "class"],
        )["id"]

        # Calculate the proportion of each class in each year
        melted = self.data.melt(var_name="year")
        melted = melted.groupby(["year", "value"]).size().reset_index(name="count")
        melted["proportion_of_total"] = (melted
            .groupby("year")["count"]
            .transform(lambda x: x / x.sum())
            .apply(lambda x: f"{x:.0%}")
        )
        all_classes = all_classes.merge(melted, left_on=["year", "class"], right_on=["year", "value"])

        if self.label_type == "class":
            all_classes["label"] = all_classes["class"].apply(lambda k: self.labels[k])
        elif self.label_type == "percent":
            all_classes["label"] = all_classes["proportion_of_total"]
        elif self.label_type == "count":
            all_classes["label"] = all_classes["count"]
        else:
            raise ValueError("Invalid label_type. Choose from 'class', 'percent', or 'count'.")

        return SankeyParameters(
            node_labels=all_classes.year,
            link_labels=df.link_label,
            node_palette=all_classes.color,
            link_palette=df.source_color,
            label=all_classes.label,
            source=df.source_id,
            target=df.target_id,
            value=df.changed,
        )

    def generate_dataframe(self) -> pd.DataFrame:
        """Convert raw sampling data to a formatted dataframe"""
        data = self.data.copy()

        if self.hide:
            hide_mask = pd.concat([(data == i).any(axis=1) for i in self.hide], axis=1).any(axis=1)
            data = data[~hide_mask]

        permutations = []
        # Get all unique class-year combinations
        for source, target in utils.pairwise(data.columns):
            permutations += list(
                zip([source] * len(data), [target] * len(data), data[source], data[target])
            )
        df = pd.DataFrame(permutations, columns=["source_year", "target_year", "source", "target"])

        # Count the unique combinations of all four fields
        df = (
            df.groupby(["source_year", "target_year", "source", "target"])
            .size()
            .reset_index()
            .rename(columns={0: "changed"})
        )
        # Count the total number of source samples in each year
        df["total"] = df.groupby(["source_year", "source"]).changed.transform(sum)
        # Calculate what percent of the source samples went into each target class
        df["proportion"] = df["changed"] / df["total"]

        # Join the class labels and colors to the class IDs
        df["source_label"] = df.source.apply(lambda k: self.labels[k])
        df["target_label"] = df.target.apply(lambda k: self.labels[k])
        df["source_color"] = df.source.apply(lambda k: self.palette[k])
        df["target_color"] = df.target.apply(lambda k: self.palette[k])

        def build_link_label(row: pd.Series) -> str:
            # Early exit in case all classes are excluded
            if row.shape[0] == 0:
                return ""

            verb = "remained" if row.source == row.target else "became"
            pct = f"{row.proportion:.0%}"
            return f"<b>{pct}</b> of <b>{row.source_label}</b> {verb} <b>{row.target_label}</b>"

        # Describe the class changes
        df["link_label"] = df.apply(build_link_label, axis=1)

        return df

    @property
    def _view_name(self):
        """When the Sankey object is displayed by IPython, render the plot"""
        return self.gui._view_name

    @property
    def _model_id(self):
        """When the Sankey object is displayed by IPython, render the plot"""
        return self.gui._model_id

    def update_layout(self, *args, **kwargs):
        """Pass layout changes to the plot. This is primarily kept for compatibility with geemap."""
        self.plot.update_layout(*args, **kwargs)

    def generate_gui(self):
        BUTTON_HEIGHT = "24px"
        BUTTON_WIDTH = "24px"

        unique_classes = self.get_sorted_classes()

        def toggle_button(button):
            button.toggle()

            class_name = button.tooltip
            class_id = [key for key in self.labels.keys() if self.labels[key] == class_name][0]

            if not button.state:
                self.hide.append(class_id)
            else:
                self.hide.remove(class_id)

            update_plot()

        def update_plot():
            """Swap new data into the plot"""
            new_plot = self.generate_plot()
            self.plot.data[0].link = new_plot.data[0].link
            self.plot.data[0].node = new_plot.data[0].node

        buttons = []
        active_classes = self.get_active_classes()
        for i in unique_classes:
            label = self.labels[i]
            on_color = self.palette[i]
            state = i in active_classes

            button = utils.ColorToggleButton(tooltip=label, on_color=on_color, state=state)
            button.layout.width = BUTTON_WIDTH
            button.layout.height = BUTTON_HEIGHT

            button.on_click(toggle_button)
            buttons.append(button)

        def reset_plot(_):
            for button in buttons:
                if not button.state:
                    button.click()

        reset_button = widgets.Button(
            icon="refresh",
            tooltip="Reset plot",
            layout=widgets.Layout(height=BUTTON_HEIGHT, width=BUTTON_WIDTH, padding="0 0 0 3px"),
        )
        reset_button.on_click(reset_plot)

        open_button = widgets.Button(
            icon="external-link",
            tooltip="Open in new tab",
            layout=widgets.Layout(height=BUTTON_HEIGHT, width=BUTTON_WIDTH, padding="0 0 0 3px"),
        )
        open_button.on_click(
            lambda _: self.plot.update_layout(width=None, height=None).show(renderer="browser")
        )

        button_box = widgets.HBox([*buttons, widgets.Label("|"), reset_button, open_button])

        gui = widgets.VBox(
            [
                self.plot,
                widgets.VBox([button_box], layout=widgets.Layout(align_items="center")),
            ]
        )

        return gui

    def generate_plot(self) -> go.Figure:
        self.df = self.generate_dataframe()
        params = self.generate_plot_parameters()

        shadow_color = "#76777a"
        label_style = f"""
            color: #fff;
            font-weight: 600;
            letter-spacing: -1px;
            text-shadow:
                0 0 4px black,
                -1px 1px 0 {shadow_color},
                1px 1px 0 {shadow_color},
                1px -1px 0 {shadow_color},
                -1px -1px 0 {shadow_color};
        """

        title_style = """
            color: #fff;
            font-weight: 900;
            word-spacing: 10px;
            letter-spacing: 3px;
            text-shadow:
                0 0 1px black,
                0 0 2px black,
                0 0 4px black;
        """

        fig = go.FigureWidget(
            data=[
                go.Sankey(
                    arrangement="snap",
                    node=dict(
                        pad=30,
                        thickness=10,
                        line=dict(color="#505050", width=1.5),
                        customdata=params.node_labels,
                        hovertemplate="<b>%{customdata}</b><extra></extra>",
                        label=[f"<span style='{label_style}'>{s}</span>" for s in params.label],
                        color=params.node_palette,
                    ),
                    link=dict(
                        source=params.source,
                        target=params.target,
                        line=dict(color="#909090", width=1),
                        value=params.value,
                        color=params.link_palette,
                        customdata=params.link_labels,
                        hovertemplate="%{customdata} <extra></extra>",
                    ),
                )
            ]
        )

        fig.update_layout(
            title_text=f"<span style='{title_style}'>{self.title}</span>" if self.title else None,
            font_size=16,
            title_x=0.5,
            paper_bgcolor="rgba(0, 0, 0, 0)",
        )

        return fig



import itertools
from collections import Counter
from typing import List

import ee
import ipywidgets as widgets


def pairwise(iterable):
    """Polyfill itertools.pairwise for pre 3.10.

    https://docs.python.org/3/library/itertools.html#itertools.pairwise
    """
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def get_shared_bands(images: List[ee.Image]) -> List[str]:
    """Get the list of bands that are shared by all images in the list.

    Args:
        img_list: List of ee.Image objects.

    Returns:
        List of band names.
    """
    band_counts = Counter(itertools.chain(*[img.bandNames().getInfo() for img in images]))
    return [band for band, count in band_counts.items() if count == len(images)]


class ColorToggleButton(widgets.Button):
    """
    The ipywidgets.ToggleButton doesn't support a `button_color` style, so this turns a standard
    ipywidgets.Button into a toggle button.
    """

    def __init__(self, *args, on_color, off_color=None, state=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.state = state
        self.on_color = on_color
        # Set an alpha value for the default off color
        self.off_color = off_color if off_color is not None else f"{on_color}20"
        current_color = on_color if state else self.off_color
        self.style.button_color = current_color
        self.layout.border = f"1px dashed {self.on_color}"

    def toggle(self):
        self.state = not self.state
        self.update()

    def update(self):
        color = self.on_color if self.state else self.off_color
        self.style.button_color = color



from typing import Dict, List, Union

import ee
import pandas as pd
import plotly.graph_objects as go

from sankee.plotting import sankify


class Dataset:
    def __init__(
        self,
        name: str,
        id: str,
        band: str,
        labels: Dict[int, str],
        palette: Dict[int, str],
        years: List[int],
        nodata: Union[None, int] = None,
    ):
        """
        Parameters
        ----------
        name : str
            The name of the dataset.
        id : str
            The :code:`system:id` of the :code:`ee.ImageCollection` representing the dataset.
        band : str
            The name of the image band that contains class values.
        labels : dict
            A dictionary matching class values to their corresponding labels.
        palette : dict
            A dictionary matching class values to their corresponding hex colors.
        years : List[int]
            All years available in this dataset.
        nodata : int
            An optional no-data value to automatically remove when plotting.
        """
        self.name = name
        self.id = id
        self.band = band
        self.labels = labels
        self.palette = palette
        self.years = years
        self.nodata = nodata

        if sorted(labels.keys()) != sorted(palette.keys()):
            raise ValueError("Labels and palette must have the same keys.")

    def __repr__(self) -> str:
        return f"<Dataset: {self.name}>"

    @property
    def keys(self) -> List[int]:
        """Return the label keys of the dataset."""
        return list(self.labels.keys())

    @property
    def df(self) -> pd.DataFrame:
        """Return a Pandas dataframe describing the dataset parameters."""
        return pd.DataFrame(
            {"id": self.keys, "label": self.labels.values(), "color": self.palette.values()}
        )

    @property
    def collection(self) -> ee.ImageCollection:
        """The :code:`ee.ImageCollection` representing the dataset."""
        return ee.ImageCollection(self.id)

    def get_year(self, year: int) -> ee.Image:
        """Get one year's image from the dataset. Set the metadata properties for visualization."""
        if year not in self.years:
            raise ValueError(
                f"This dataset does not include year `{year}`. Choose from {self.years}."
            )

        img = self.collection.filterDate(str(year), str(year + 1)).first()
        img = self.set_visualization_properties(img)

        if self.nodata is not None:
            img = img.updateMask(img.neq(self.nodata))

        return img.select(self.band)

    def set_visualization_properties(self, image: ee.Image) -> ee.Image:
        """Set the properties used by Earth Engine to automatically assign a palette to an image
        from this dataset."""
        return image.set(
            f"{self.band}_class_values",
            list(self.labels.keys()),
            f"{self.band}_class_palette",
            [c.replace("#", "") for c in self.palette.values()],
        )

    def list_years(self) -> ee.List:
        """Get an ee.List of all years in the collection."""
        return (
            self.collection.aggregate_array("system:time_start")
            .map(lambda ms: ee.Date(ms).get("year"))
            .distinct()
        )

    def sankify(
        self,
        years: List[int],
        region: ee.Geometry,
        max_classes: Union[None, int] = None,
        n: int = 500,
        title: Union[str, None] = None,
        scale: Union[int, None] = None,
        seed: int = 0,
        exclude: None = None,
        label_type: str = "class",
    ) -> go.Figure:
        """
        Generate an interactive Sankey plot showing land cover change over time from a series of
        years in the dataset.

        Parameters
        ----------
        years : List[int]
            The years to include in the plot. Select at least two unique years.
        region : ee.Geometry
            A region to generate samples within. The region must overlap all images.
        exclude : list[int], default None
            An optional list of pixel values to exclude from the plot. Excluded values must be raw
            pixel values rather than class labels. This can be helpful if the region is dominated by
            one or more unchanging classes and the goal is to visualize changes in smaller classes.
            No-data classes are always excluded automatically.
        max_classes : int, default None
            If a value is provided, small classes will be removed until max_classes remain. Class
            size is calculated based on total times sampled in the time series.
        n : int, default 500
            The number of sample points to randomly generate for characterizing all images. More
            samples will provide more representative data but will take longer to process.
        title : str, default None
            An optional title that will be displayed above the Sankey plot.
        scale : int, default None
            The scale in image units to perform sampling at. If none is provided, GEE will attempt
            to use the image's nominal scale, which may cause errors depending on the image
            projection.
        seed : int, default 0
            The seed value used to generate repeatable results during random sampling.
        exclude : None
            Unused parameter that will be removed in a future release.
        label_type : str, default "class"
            The type of label to display for each link, one of "class", "percent", or "count". Selecting
            "class" will use the class label, "percent" will use the proportion of sampled pixels in each
            class, and "count" will use the number of sampled pixels in each class.

        Returns
        -------
        plotly.graph_objs._figure.Figure
            An interactive Sankey plot.
        """
        if len(years) < 2:
            raise ValueError("Select at least two years.")
        if len(set(years)) != len(years):
            raise ValueError("Duplicate years found. Make sure all years are unique.")
        years = sorted(years)

        imgs = [self.get_year(year) for year in years]

        labels = self.labels.copy()
        palette = self.palette.copy()
        if self.nodata is not None:
            labels.pop(self.nodata)
            palette.pop(self.nodata)

        return sankify(
            image_list=imgs,
            label_list=years,
            labels=labels,
            band=self.band,
            palette=palette,
            region=region,
            max_classes=max_classes,
            n=n,
            title=title,
            scale=scale,
            seed=seed,
            label_type=label_type,
        )


class LCMS_Dataset(Dataset):
    def get_year(self, year: int) -> ee.Image:
        """Get one year's image from the dataset. LCMS splits up each year into two images: CONUS
        and SEAK. This merges those into a single image."""
        super().get_year(year)

        collection = self.collection.filter(ee.Filter.eq("year", year))
        merged = collection.mosaic().select(self.band).clip(collection.geometry())

        props = collection.first().propertyNames().remove("study_area")
        merged = ee.Image(ee.Element.copyProperties(merged, collection.first(), props))

        merged = merged.setDefaultProjection("EPSG:5070")
        return merged


class CCAP_Dataset(Dataset):
    def get_year(self, year: int) -> ee.Image:
        """Get one year's image from the dataset. C-CAP splits up each year into multiple images,
        so merge those and set the class value and palette metadata to allow automatic
        visualization."""
        super().get_year(year)

        imgs = self.collection.filterDate(str(year), str(year + 1))

        img = (
            imgs.mosaic()
            .set(
                {
                    "system:time_start": imgs.first().get("system:time_start"),
                    "system:time_end": imgs.first().get("system:time_end"),
                }
            )
            .clip(imgs.geometry())
            .setDefaultProjection("EPSG:5070")
        )

        img = self.set_visualization_properties(img)

        return img


LCMS_LU = LCMS_Dataset(
    name="LCMS LU - Land Change Monitoring System Land Use",
    id="USFS/GTAC/LCMS/v2021-7",
    band="Land_Use",
    labels={
        1: "Agriculture",
        2: "Developed",
        3: "Forest",
        4: "Non-Forest Wetland",
        5: "Other",
        6: "Rangeland or Pasture",
        7: "No Data",
    },
    palette={
        1: "#efff6b",
        2: "#ff2ff8",
        3: "#1b9d0c",
        4: "#97ffff",
        5: "#a1a1a1",
        6: "#c2b34a",
        7: "#1B1716",
    },
    years=tuple(range(1985, 2022)),
    nodata=7,
)

# https://developers.google.com/earth-engine/datasets/catalog/USFS_GTAC_LCMS_v2020-5
LCMS_LC = LCMS_Dataset(
    name="LCMS LC - Land Change Monitoring System Land Cover",
    id="USFS/GTAC/LCMS/v2021-7",
    band="Land_Cover",
    labels={
        1: "Trees",
        2: "Tall Shrubs & Trees Mix",
        3: "Shrubs & Trees Mix",
        4: "Grass/Forb/Herb & Trees Mix",
        5: "Barren & Trees Mix",
        6: "Tall Shrubs",
        7: "Shrubs",
        8: "Grass/Forb/Herb & Shrubs Mix",
        9: "Barren & Shrubs Mix",
        10: "Grass/Forb/Herb",
        11: "Barren & Grass/Forb/Herb Mix",
        12: "Barren or Impervious",
        13: "Snow or Ice",
        14: "Water",
        15: "No Data",
    },
    palette={
        1: "#005e00",
        2: "#008000",
        3: "#00cc00",
        4: "#b3ff1a",
        5: "#99ff99",
        6: "#b30088",
        7: "#e68a00",
        8: "#ffad33",
        9: "#ffe0b3",
        10: "#ffff00",
        11: "#AA7700",
        12: "#d3bf9b",
        13: "#ffffff",
        14: "#4780f3",
        15: "#1B1716",
    },
    years=tuple(range(1985, 2022)),
    nodata=15,
)

NLCD = Dataset(
    name="NLCD - National Land Cover Database",
    id="USGS/NLCD_RELEASES/2019_REL/NLCD",
    band="landcover",
    labels={
        1: "No data",
        11: "Open water",
        12: "Perennial ice/snow",
        21: "Developed, open space",
        22: "Developed, low intensity",
        23: "Developed, medium intensity",
        24: "Developed, high intensity",
        31: "Barren land (rock/sand/clay)",
        41: "Deciduous forest",
        42: "Evergreen forest",
        43: "Mixed forest",
        51: "Dwarf scrub",
        52: "Shrub/scrub",
        71: "Grassland/herbaceous",
        72: "Sedge/herbaceous",
        73: "Lichens",
        74: "Moss",
        81: "Pasture/hay",
        82: "Cultivated crops",
        90: "Woody wetlands",
        95: "Emergent herbaceous wetlands",
    },
    palette={
        1: "#000000",
        11: "#466b9f",
        12: "#d1def8",
        21: "#dec5c5",
        22: "#d99282",
        23: "#eb0000",
        24: "#ab0000",
        31: "#b3ac9f",
        41: "#68ab5f",
        42: "#1c5f2c",
        43: "#b5c58f",
        51: "#af963c",
        52: "#ccb879",
        71: "#dfdfc2",
        72: "#d1d182",
        73: "#a3cc51",
        74: "#82ba9e",
        81: "#dcd939",
        82: "#ab6c28",
        90: "#b8d9eb",
        95: "#6c9fb8",
    },
    years=(2001, 2004, 2006, 2008, 2011, 2013, 2016, 2019),
    nodata=1,
)

# Kept for backwards compatibility with older `geemap` versions
NLCD2016 = NLCD

MODIS_LC_TYPE1 = Dataset(
    name="MCD12Q1 - MODIS Global Land Cover Type 1",
    id="MODIS/006/MCD12Q1",
    band="LC_Type1",
    labels={
        1: "Evergreen conifer forest",
        2: "Evergreen broadleaf forest",
        3: "Deciduous conifer forest",
        4: "Deciduous broadleaf forest",
        5: "Mixed forest",
        6: "Closed shrubland",
        7: "Open shrubland",
        8: "Woody savanna",
        9: "Savanna",
        10: "Grassland",
        11: "Permanent wetland",
        12: "Cropland",
        13: "Urban",
        14: "Cropland and natural vegetation",
        15: "Permanent snow and ice",
        16: "Barren",
        17: "Water",
    },
    palette={
        1: "#086a10",
        2: "#dcd159",
        3: "#54a708",
        4: "#78d203",
        5: "#009900",
        6: "#c6b044",
        7: "#dcd159",
        8: "#dade48",
        9: "#fbff13",
        10: "#b6ff05",
        11: "#27ff87",
        12: "#c24f44",
        13: "#a5a5a5",
        14: "#ff6d4c",
        15: "#69fff8",
        16: "#f9ffa4",
        17: "#1c0dff",
    },
    years=tuple(range(2001, 2021)),
)

# https://developers.google.com/earth-engine/datasets/catalog/MODIS_006_MCD12Q1
MODIS_LC_TYPE2 = Dataset(
    name="MCD12Q1 - MODIS Global Land Cover Type 2",
    id="MODIS/006/MCD12Q1",
    band="LC_Type2",
    labels={
        0: "Water",
        1: "Evergreen conifer forest",
        2: "Evergreen broadleaf forest",
        3: "Deciduous conifer forest",
        4: "Deciduous broadleaf forest",
        5: "Mixed forest",
        6: "Closed shrubland",
        7: "Open shrubland",
        8: "Woody savanna",
        9: "Savanna",
        10: "Grassland",
        11: "Permanent wetland",
        12: "Cropland",
        13: "Urban",
        14: "Cropland and natural vegetation",
        15: "Barren",
    },
    palette={
        0: "#1c0dff",
        1: "#05450a",
        2: "#086a10",
        3: "#54a708",
        4: "#78d203",
        5: "#009900",
        6: "#c6b044",
        7: "#dcd159",
        8: "#dade48",
        9: "#fbff13",
        10: "#b6ff05",
        11: "#27ff87",
        12: "#c24f44",
        13: "#a5a5a5",
        14: "#ff6d4c",
        15: "#f9ffa4",
    },
    years=tuple(range(2001, 2021)),
)

# https://developers.google.com/earth-engine/datasets/catalog/MODIS_006_MCD12Q1
MODIS_LC_TYPE3 = Dataset(
    name="MCD12Q1 - MODIS Global Land Cover Type 3",
    id="MODIS/006/MCD12Q1",
    band="LC_Type3",
    labels={
        0: "Water",
        1: "Grassland",
        2: "Shrubland",
        3: "Crops",
        4: "Savannas",
        5: "Evergreen broadleaf",
        6: "Deciduous broadleaf",
        7: "Evergreen conifer",
        8: "Deciduous conifer",
        9: "Barren",
        10: "Urban",
    },
    palette={
        0: "#1c0dff",
        1: "#b6ff05",
        2: "#dcd159",
        3: "#c24f44",
        4: "#fbff13",
        5: "#086a10",
        6: "#78d203",
        7: "#05450a",
        8: "#54a708",
        9: "#f9ffa4",
        10: "#a5a5a5",
    },
    years=tuple(range(2001, 2021)),
)

# https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_Landcover_100m_Proba-V-C3_Global
CGLS_LC100 = Dataset(
    name="CGLS - Copernicus Global Land Cover",
    id="COPERNICUS/Landcover/100m/Proba-V-C3/Global",
    band="discrete_classification",
    labels={
        0: "Unknown",
        20: "Shrubs",
        30: "Herbaceous vegetation",
        40: "Cultivated",
        50: "Urban",
        60: "Bare",
        70: "Snow and ice",
        80: "Water body",
        90: "Herbaceous wetland",
        100: "Moss and lichen",
        111: "Closed forest, evergreen conifer",
        112: "Closed forest, evergreen broad leaf",
        113: "Closed forest, deciduous conifer",
        114: "Closed forest, deciduous broad leaf",
        115: "Closd forest, mixed",
        116: "Closed forest, other",
        121: "Open forest, evergreen conifer",
        122: "Open forest, evergreen broad leaf",
        123: "Open forest, deciduous conifer",
        124: "Open forest, deciduous broad leaf",
        125: "Open forest, mixed",
        126: "Open forest, other",
        200: "Ocean",
    },
    palette={
        0: "#282828",
        20: "#FFBB22",
        30: "#FFFF4C",
        40: "#F096FF",
        50: "#FA0000",
        60: "#B4B4B4",
        70: "#F0F0F0",
        80: "#0032C8",
        90: "#0096A0",
        100: "#FAE6A0",
        111: "#58481F",
        112: "#009900",
        113: "#70663E",
        114: "#00CC00",
        115: "#4E751F",
        116: "#007800",
        121: "#666000",
        122: "#8DB400",
        123: "#8D7400",
        124: "#A0DC00",
        125: "#929900",
        126: "#648C00",
        200: "#000080",
    },
    years=tuple(range(2015, 2020)),
    nodata=0,
)

# https://samapriya.github.io/awesome-gee-community-datasets/projects/ccap_mlc/
CCAP_LC30 = CCAP_Dataset(
    name="C-CAP - NOAA Coastal Change Analysis Program 30m",
    id="projects/sat-io/open-datasets/NOAA/ccap_30m",
    band="b1",
    labels={
        1: "Unclassified (Cloud,Shadow etc)",
        2: "Impervious",
        3: "Developed Open Space",
        4: "Developed Open Space",
        5: "Developed Open Space",
        6: "Cultivated Land",
        7: "Pasture/Hay",
        8: "Grassland/Herbaceous",
        9: "Deciduous Forest",
        10: "Evergreen Forest",
        11: "Mixed Forest",
        12: "Scrub/Shrub",
        13: "Palustrine Forested Wetland",
        14: "Palustrine Scrub/Shrub Wetland",
        15: "Palustrine Emergent Wetland",
        16: "Estuarine Forested Wetland",
        17: "Estuarine Scrub/Shrub Wetland",
        18: "Estuarine Emergent Wetland",
        19: "Unconsolidated Shore",
        20: "Bare Land",
        21: "Open Water",
        22: "Palustrine Aquatic Bed",
        23: "Estuarine Aquatic Bed",
        24: "Tundra",
        25: "Snow/Ice",
    },
    palette={
        1: "#000000",
        2: "#f2f2f2",
        3: "#a899a8",
        4: "#8e757c",
        5: "#c1cc38",
        6: "#542100",
        7: "#c1a04f",
        8: "#f2ba87",
        9: "#00f200",
        10: "#003a00",
        11: "#07a03a",
        12: "#6d6d00",
        13: "#005b5b",
        14: "#f26d00",
        15: "#f200f2",
        16: "#3d003d",
        17: "#6d006d",
        18: "#af00af",
        19: "#00f2f2",
        20: "#f2f200",
        21: "#000077",
        22: "#0000f2",
        23: "#161616",
        24: "#161616",
        25: "#191919",
    },
    years=(1975, 1985, 1992, 1996, 2001, 2006, 2010, 2016),
    nodata=1,
)

# https://samapriya.github.io/awesome-gee-community-datasets/projects/ca_lc/
CA_FOREST_LC = Dataset(
    name="Canada Forested Ecosystem Land Cover",
    id="projects/sat-io/open-datasets/CA_FOREST_LC_VLCE2",
    band="b1",
    labels={
        0: "Unclassified",
        20: "Water",
        31: "Snow/Ice",
        32: "Rock/Rubble",
        33: "Exposed/Barren land",
        40: "Bryoids",
        50: "Shrubs",
        80: "Wetland",
        81: "Wetland-treed",
        100: "Herbs",
        210: "Coniferous",
        220: "Broadleaf",
        230: "Mixedwood",
    },
    palette={
        0: "#686868",
        20: "#3333ff",
        31: "#ccffff",
        32: "#cccccc",
        33: "#996633",
        40: "#ffccff",
        50: "#ffff00",
        80: "#993399",
        81: "#9933cc",
        100: "#ccff33",
        210: "#006600",
        220: "#00cc00",
        230: "#cc9900",
    },
    years=tuple(range(1984, 2020)),
    nodata=0,
)

LCMAP = Dataset(
    name="LCMAP - Landscape Change Monitoring, Assessment, and Projection",
    id="projects/sat-io/open-datasets/LCMAP/LCPRI",
    band="b1",
    labels={
        1: "Developed",
        2: "Cropland",
        3: "Grass/Shrub",
        4: "Tree Cover",
        5: "Water",
        6: "Wetland",
        7: "Ice/Snow",
        8: "Barren",
    },
    palette={
        1: "#E60000",
        2: "#A87000",
        3: "#E3E3C2",
        4: "#1D6330",
        5: "#476BA1",
        6: "#BAD9EB",
        7: "#FFFFFF",
        8: "#B3B0A3",
    },
    years=tuple(range(1985, 2021)),
)

CORINE = Dataset(
    name="CORINE - Coordination of Information on the Environment",
    id="COPERNICUS/CORINE/V20/100m",
    band="landcover",
    # Note: Labels are shortened for display
    labels={
        111: "Continuous urban",
        112: "Discontinuous urban",
        121: "Industrial/Commercial",
        122: "Road/Rail",
        123: "Ports",
        124: "Airports",
        131: "Mines",
        132: "Dump sites",
        133: "Construction",
        141: "Green urban",
        142: "Sport/Leisure facilities",
        211: "Non-irrigated arable",
        212: "Permanently irrigated",
        213: "Rice fields",
        221: "Vineyards",
        222: "Fruit plantations",
        223: "Olive groves",
        231: "Pastures",
        241: "Annual crops",
        242: "Complex cultivation",
        243: "Agriculture/Natural",
        244: "Agro-forestry",
        311: "Broad-leaved forest",
        312: "Coniferous forest",
        313: "Mixed forest",
        321: "Natural grasslands",
        322: "Moors and heathland",
        323: "Sclerophyllous vegetation",
        324: "Transitional woodland-shrub",
        331: "Beaches, dunes, sand",
        332: "Bare rocks",
        333: "Sparsely vegetated",
        334: "Burnt",
        335: "Glaciers/Perpetual snow",
        411: "Inland marshes",
        412: "Peat bogs",
        421: "Salt marshes",
        422: "Salines",
        423: "Intertidal flats",
        511: "Water courses",
        512: "Water bodies",
        521: "Coastal lagoons",
        522: "Estuaries",
        523: "Sea/Ocean",
    },
    palette={
        111: "#E6004D",
        112: "#FF0000",
        121: "#CC4DF2",
        122: "#CC0000",
        123: "#E6CCCC",
        124: "#E6CCE6",
        131: "#A600CC",
        132: "#A64DCC",
        133: "#FF4DFF",
        141: "#FFA6FF",
        142: "#FFE6FF",
        211: "#FFFFA8",
        212: "#FFFF00",
        213: "#E6E600",
        221: "#E68000",
        222: "#F2A64D",
        223: "#E6A600",
        231: "#E6E64D",
        241: "#FFE6A6",
        242: "#FFE64D",
        243: "#E6CC4D",
        244: "#F2CCA6",
        311: "#80FF00",
        312: "#00A600",
        313: "#4DFF00",
        321: "#CCF24D",
        322: "#A6FF80",
        323: "#A6E64D",
        324: "#A6F200",
        331: "#E6E6E6",
        332: "#CCCCCC",
        333: "#CCFFCC",
        334: "#000000",
        335: "#A6E6CC",
        411: "#A6A6FF",
        412: "#4D4DFF",
        421: "#CCCCFF",
        422: "#E6E6FF",
        423: "#A6A6E6",
        511: "#00CCF2",
        512: "#80F2E6",
        521: "#00FFA6",
        522: "#A6FFE6",
        523: "#E6F2FF",
    },
    years=(1986, 1999, 2005, 2011, 2017),
)


datasets = [
    LCMS_LC,
    LCMS_LU,
    NLCD,
    MODIS_LC_TYPE1,
    MODIS_LC_TYPE2,
    MODIS_LC_TYPE3,
    CGLS_LC100,
    CCAP_LC30,
    CA_FOREST_LC,
    LCMAP,
    CORINE,
]



from typing import List, Tuple, Union

import ee
import pandas as pd

from sankee import utils


def generate_sample_data(
    *,
    image_list: List[ee.Image],
    image_labels: List[str],
    region: ee.Geometry,
    band: str,
    n: int = 500,
    scale: Union[None, int] = None,
    seed: int = 0,
    include: Union[None, List[int]] = None,
    max_classes: Union[None, int] = None,
) -> Tuple[pd.DataFrame, ee.FeatureCollection]:
    """Take a list of images extract image values to each to random points. The image values will be
    stored in a property based on the image label. Then, the samples will be returned as a formated
    dataframe with one column for each image and one row for each sample point.
    """

    def extract_values_at_point(pt):
        for img, label in zip(image_list, image_labels):
            cover = img.reduceRegion(
                reducer=ee.Reducer.first(), geometry=pt.geometry(), scale=scale
            ).get(band)
            pt = ee.Feature(pt).set(label, cover)

        return pt

    points = ee.FeatureCollection.randomPoints(region=region, points=n, seed=seed)
    samples = points.map(extract_values_at_point)

    try:
        features = [feat["properties"] for feat in samples.toList(samples.size()).getInfo()]
    except ee.EEException as e:
        if band in str(e):
            shared_bands = utils.get_shared_bands(image_list)
            raise ValueError(
                f"The band `{band}` was not found in all images. Choose from " f"{shared_bands}"
            ) from None
        elif "'count' must be positive" in str(e):
            raise ValueError("No points were sampled. Make sure to pass a 2D `region`.") from None
        else:
            raise e

    data = pd.DataFrame.from_dict(features).dropna().astype(int)

    for image in image_labels:
        if image not in data.columns:
            raise ValueError(
                f"Valid samples were not found for image `{image}`. Check that the"
                " image overlaps the sampling region."
            )

    # EE data gets sorted alpha, so re-sort columns to the image order. Run this after the column
    # check because this can add columns that were missing.
    data = data.reindex(image_labels, axis=1)
    
    if include is not None:
        data = data[data.isin(include).all(axis=1)]

    if max_classes is not None:
        class_counts = data.melt().value.value_counts()
        keep_classes = class_counts[:max_classes].index.tolist()
        data = data[data.isin(keep_classes).all(axis=1)]

    return data, samples



from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "earthengine-api>=0.1.230",
    "numpy",
    "pandas",
    "plotly>=5.2.2",
    "ipywidgets",
]
doc_requirements = ["nbsphinx", "sphinx", "sphinx_rtd_theme"]
test_requirements = ["pytest", "coverage", "pytest-cov", "httplib2"]
dev_requirements = (
    [
        "pre-commit",
        "mypy",
        "black",
        "isort",
        "bumpversion",
        "twine",
    ]
    + doc_requirements
    + test_requirements
)

extras_requirements = {
    "doc": doc_requirements,
    "dev": dev_requirements,
    "test": test_requirements,
}


setup(
    name="sankee",
    version="0.2.1",
    description="Visualize classified time series data with interactive Sankey plots in Google Earth Engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aazuspan/sankee",
    author="Aaron Zuspan",
    author_email="aa.zuspan@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="sankey land cover visualization",
    packages=find_packages(include=["sankee", "sankee."]),
    python_requires=">=3.7",
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require=extras_requirements,
    project_urls={
        "Bug Reports": "https://github.com/aazuspan/sankee/issues",
        "Source": "https://github.com/aazuspan/sankee/",
    },
)
